/*********************************************************************
Project	:	GUSI				-	Grand Unified Socket Interface
File		:	GUSI_P.h			-	Private stuff
Author	:	Matthias Neeracher
Language	:	MPW C/C++

$Log: GUSI_P.h,v $
Revision 1.1  1994/02/25  02:57:01  neeri
Initial revision

Revision 0.22  1993/07/17  00:00:00  neeri
GUSIRingBuffer::proc -> defproc

Revision 0.21  1993/07/17  00:00:00  neeri
GUSIO_MAX_DOMAIN -> AF_MAX

Revision 0.20  1993/06/27  00:00:00  neeri
Socket::{pre,post}_select

Revision 0.19  1993/06/27  00:00:00  neeri
Socket::ftruncate

Revision 0.18  1993/02/09  00:00:00  neeri
Socket::lurking, Socket::lurkdescr

Revision 0.17  1993/01/31  00:00:00  neeri
GUSIConfiguration::daemon

Revision 0.16  1993/01/17  00:00:00  neeri
Destructors for Socketdomain

Revision 0.15  1993/01/17  00:00:00  neeri
SAFESPIN

Revision 0.14  1993/01/03  00:00:00  neeri
GUSIConfig

Revision 0.13  1992/09/24  00:00:00  neeri
Include GUSIRsrc_P.h

Revision 0.12  1992/09/13  00:00:00  neeri
SPINVOID didn't return

Revision 0.11  1992/08/30  00:00:00  neeri
AppleTalkIdentity()

Revision 0.10  1992/08/03  00:00:00  neeri
RingBuffer

Revision 0.9  1992/07/30  00:00:00  neeri
Initializer Features

Revision 0.8  1992/07/26  00:00:00  neeri
UnixSockets.choose()

Revision 0.7  1992/07/13  00:00:00  neeri
Make AppleTalkSockets global

Revision 0.6  1992/06/27  00:00:00  neeri
choose(), hasNewSF

Revision 0.5  1992/06/07  00:00:00  neeri
Feature

Revision 0.4  1992/05/21  00:00:00  neeri
Implemented select()

Revision 0.3  1992/04/19  00:00:00  neeri
C++ rewrite

Revision 0.2  1992/04/18  00:00:00  neeri
changed read/write/send/recv dispatchers

Revision 0.1  1992/04/18  00:00:00  neeri
ppc Domain

*********************************************************************/

#include <GUSI.h>
#include <GUSIRsrc_P.h>
#include <TFileSpec.h>

extern "C" {

#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <Memory.h>
#include <GestaltEqu.h>
#include <Traps.h>
#include <AppleEvents.h>

int 		GUSI_error(int err);
void *	GUSI_error_nil(int err);
}

#pragma segment GUSI

#define GUSI_MAX_DOMAIN			AF_MAX
#define DEFAULT_BUFFER_SIZE	4096

/*
 *	In use and shutdown status.
 */
#define	SOCK_STATUS_USED		0x1		/* Used socket table entry */
#define	SOCK_STATUS_NOREAD	0x2		/* No more reading allowed from socket */
#define	SOCK_STATUS_NOWRITE	0x4		/* No more writing allowed to socket */

/*
 *	Socket connection states.
 */
#define	SOCK_STATE_NO_STREAM		0	/* Socket doesn't have a MacTCP stream yet */
#define	SOCK_STATE_UNCONNECTED	1	/* Socket is unconnected. */
#define	SOCK_STATE_LISTENING		2	/* Socket is listening for connection. */
#define	SOCK_STATE_LIS_CON		3	/* Socket is in transition from listen to connected. */
#define	SOCK_STATE_CONNECTING	4	/* Socket is initiating a connection. */
#define	SOCK_STATE_CONNECTED		5	/* Socket is connected. */
#define	SOCK_STATE_CLOSING      6	/* Socket is closing */

#define		min(a,b)				( (a) < (b) ? (a) : (b))
#define		max(a,b)				( (a) > (b) ? (a) : (b))

extern GUSISpinFn GUSISpin;
extern "C" int GUSIDefaultSpin(spin_msg, long);

/* SPIN returns a -1 on user cancel for fn returning integers */
#define		SPIN(cond,mesg,param)							\
					do {												\
						if (GUSISpin)							\
							if ((*GUSISpin)(mesg,param))	\
								return GUSI_error(EINTR);		\
					} while(cond)

/* SPINP returns a NULL on user cancel, for fn returning pointers */				
#define		SPINP(cond,mesg,param)							\
					do {												\
						if (GUSISpin)							\
							if ((*GUSISpin)(mesg,param)) {\
								GUSI_error(EINTR);				\
								return NULL;						\
							}											\
					} while(cond)

/* SPINVOID just returns on user cancel, for fn returning void */				
#define		SPINVOID(cond,mesg,param)						\
					do {												\
						if (GUSISpin)								\
							if ((*GUSISpin)(mesg,param)) {	\
								GUSI_error(EINTR);				\
								return;								\
							}											\
					} while(cond)
					
/* SAFESPIN doesn't return, you have to check errno */				
#define		SAFESPIN(cond,mesg,param)						\
					do {												\
						if (GUSISpin)								\
							if ((*GUSISpin)(mesg,param)) {	\
								GUSI_error(EINTR);				\
								break;								\
							} else {									\
								errno = 0;							\
							}											\
						else											\
							errno = 0;								\
					} while(cond)
					
class SocketTable;

class Socket {
	friend class SocketTable;
	
	short			refCount;
protected:
	Boolean		lurking;
	char			lurkDescr;
	
					Socket();
public:
	virtual int	bind(void * name, int namelen);
	virtual int connect(void * address, int addrlen);
	virtual int listen(int qlen);
	virtual Socket * accept(void * address, int * addrlen);
	virtual int	read(void * buffer, int buflen);
	virtual int write(void * buffer, int buflen);
	virtual int recvfrom(void * buffer, int buflen, int flags, void * from, int * fromlen);
	virtual int sendto(void * buffer, int buflen, int flags, void * to, int tolen);
	virtual int getsockname(void * name, int * namelen);
	virtual int getpeername(void * name, int * namelen);
	virtual int getsockopt(int level, int optname, void *optval, int * optlen);
	virtual int setsockopt(int level, int optname, void *optval, int optlen);
	virtual int	fcntl(unsigned int cmd, int arg);
	virtual int	ioctl(unsigned int request, void *argp);
	virtual int	fstat(struct stat * buf);
	virtual long lseek(long offset, int whence);
	virtual int ftruncate(long offset);
	virtual int	isatty();
	virtual int shutdown(int how);
	virtual void pre_select(Boolean wantRead, Boolean wantWrite, Boolean wantExcept);
	virtual int select(Boolean * canRead, Boolean * canWrite, Boolean * exception);
	virtual void post_select(Boolean wantRead, Boolean wantWrite, Boolean wantExcept);
	virtual 		~Socket();
};

class SocketDomain {
	static SocketDomain *	domains[GUSI_MAX_DOMAIN];
protected:
	SocketDomain(int domain);
	virtual ~SocketDomain();
public:
	inline static SocketDomain *	Domain(int domain);
	
	// Override *one* of the following two
	
	virtual Socket * open(const char * filename, int oflag);
	virtual Socket * socket(int type, short protocol);
	
	// Optionally define the following
	
	virtual int choose(
						int 		type, 
						char * 	prompt, 
						void * 	constraint,		
						int 		flags,
 						void * 	name, 
						int * 	namelen);
	
	// Never override the following
	
	virtual void DontStrip();
};

class DeviceSocketDomain : public SocketDomain {
protected:
	friend class FileSocketDomain;
	
	static DeviceSocketDomain *	firstDev;
	static const Socket *			TryNextDevice;
	DeviceSocketDomain *				nextDev;

	DeviceSocketDomain(int domain);
};

class SocketTable {
	Socket *	sockets[GUSI_MAX_FD];
public:
	SocketTable();
	~SocketTable();
	
	int		Install(Socket * sock, int start = 0);
	int		Remove(int fd);
	Socket * operator[](int fd);
	
	void		Possess(int descr, Socket * sock);	// Install a daemon
};

class FileSocketDomain : public SocketDomain {
	friend SocketTable::SocketTable();
	
	virtual Socket * stdopen(int fd);
public:
	FileSocketDomain()	:	SocketDomain(AF_FILE)	{	}

	virtual Socket * open(const char * filename, int oflag);
	virtual int choose(
						int 		type, 
						char * 	prompt, 
						void * 	constraint,		
						int 		flags,
 						void * 	name, 
						int * 	namelen);
};

struct GUSISuffix {
	char 		suffix[4];
	OSType	suffType;
	OSType	suffCreator;
};

struct GUSIConfiguration {
	OSType			defaultType;
	OSType			defaultCreator;
	
	char				autoSpin;
	
	unsigned 		noChdir 		: 1;	// Set current directory without chdir()
	unsigned 		accurStat	: 1;	// Return # of subdirectories + 2 in st_nlink
	unsigned 		tcpDaemon	: 1;	// Works as a TCP daemon ?
	unsigned 		udpDaemon	: 1;	// Works as an UDP daemon ?
	unsigned 		filler2		: 4;
	
	OSType			version;
	short				numSuffices;
	GUSISuffix *	suffices;
	
	GUSIConfiguration();
	
	void SetDefaultFType(const TFileSpec & name) const;
	void DoAutoSpin() const;
	Boolean IsDaemon() const;
private:
	static Boolean firstTime;
	static short	we;
};

extern const GUSIConfiguration	GUSIConfig;
extern SocketTable					Sockets;
extern FileSocketDomain 			FileSockets;

typedef pascal OSErr (*OSErrInitializer)();
typedef pascal void  (*voidInitializer)();

class Feature {
	Boolean	good;
public:
	Feature(short trapNum, TrapType tTyp);
	Feature(OSType type, long value);
	Feature(OSType type, long mask, long value);
	Feature(const Feature & precondition, OSErrInitializer init);
	Feature(OSErrInitializer init);
	Feature(const Feature & precondition, voidInitializer init);
	Feature(voidInitializer init);
	Feature(const Feature & cond1, const Feature & cond2);

	operator void*() const {	return (void *) good;	}
};

extern Feature hasMakeFSSpec;
extern Feature hasAlias;
extern Feature	hasWNE;
extern Feature hasNewSF;
extern Feature hasProcessMgr;
extern Feature hasCRM;
extern Feature hasCTB;
extern Feature hasStdNBP;
extern Feature hasCM;
extern Feature hasFT;
extern Feature hasTM;
extern Feature	hasPPC;
extern Feature hasRevisedTimeMgr;

class ScattGath	{
	Handle			scratch;
protected:
	void *			buf;
	int						len;
	int						count;
	const struct iovec *	io;

	ScattGath(const struct iovec *iov, int cnt);
	virtual ~ScattGath();
public:
	void *			buffer()			{	return buf;			}
	int				buflen()			{	return len;			}
	int				length(int l)	{	return len = l;	}
	operator void *()					{	return buf;			}
};

class Scatterer : public ScattGath {
public:
	Scatterer(const struct iovec *iov, int count);
	virtual ~Scatterer();
};

class Gatherer : public ScattGath {
public:
	Gatherer(const struct iovec *iov, int count);
	virtual ~Gatherer();
};

typedef pascal void (*Deferred)(void *);

class RingBuffer {
	// Valid bytes are between consume and produce
	// Free bytes are between produce and consume
	// bytes between endbuf-spare and endbuf are neither
	Ptr		buffer;
	Ptr		endbuf;
	Ptr 		consume;
	Ptr		produce;
	u_short	free;
	u_short	valid;
	u_short	spare;
	Boolean	lock;
	Deferred	defproc;
	void *	arg;
	
public:
				RingBuffer(u_short bufsiz);
				~RingBuffer();
	
	Ptr		Producer(long & len);			//	Find continuous memory for producer
	Ptr		Consumer(long & len);			//	Find continuous memory for consumer
	void		Validate(long len);				// Validate this, unallocate rest
	void 		Invalidate(long len);
	void		Produce(Ptr from, long & len);//	Allocate, copy & validate
	void		Consume(Ptr to, long & len);	// Copy & invalidate
	
	long		Free()								{ return free;									}		
	long		Valid()								{ return valid;								}
	
	void 		Defer()								{ lock = true;									}
	void 		Undefer()							{ lock = false; if (defproc) defproc(arg);}
	Boolean	Locked()								{ return lock;									}
	void		Later(Deferred def, void * ar){ defproc = def; arg = ar;					}
	
	operator void *()								{ return buffer;								}
};

Boolean GUSIInterrupt();

Boolean CopyIconFamily(short srcResFile, short srcID, short dstResFile, short dstID);

pascal OSErr PPCInit_P();

OSErr AppleTalkIdentity(short & net, short & node);

pascal OSErr GetOffMyCloud(AppleEvent*, AppleEvent*, long); // We don't want this aevt
